//: DatabaseImpl.c 

#include "com_hp_turboimage_DatabaseImpl.h"

/* function to throw exceptions based on TurboIMAGE errors */
void ThrowException(JNIEnv *env, char * intrinsic, void *CStatus){
   char   message[100];
   short  msglen;
   jclass IEClass;

   /* first report & clear any pending exception */
   (*env)->ExceptionDescribe(env);
   (*env)->ExceptionClear(env);

   /* now see what our problem is */
   dberror(CStatus, message, &msglen);
   message[msglen] = '\0'; /* needs null terminator */

   /* and throw an exception to report it */
   IEClass=(*env)->FindClass(env, "com/hp/turboimage/IMAGEException");
   (*env)->ThrowNew(env, IEClass, message);
}

//**********************************************************
//********** TurboIMAGE Database Implementation ************
//**********************************************************

/* DBOPEN */
JNIEXPORT void JNICALL
Java_com_hp_turboimage_DatabaseImpl_DBOPEN(JNIEnv    *env,
                                       jobject    this,
                                       jbyteArray j_basename,
                                       jstring    j_password,
                                       jint       j_mode,
                                       jobject    j_status) {

   // Get JNI fields into C types
   jbyte         *base = (*env)->GetByteArrayElements(env, j_basename, 0);
   const char    *pass = (*env)->GetStringUTFChars(env, j_password, 0);
   jclass   c_DBStatus = (*env)->GetObjectClass(env, j_status);
   jfieldID f_halfword = (*env)->GetFieldID(env, c_DBStatus, "halfword", "[S");
   jobject  p_halfword = (*env)->GetObjectField(env, j_status, f_halfword);
   jshort  *a_halfword = (*env)->GetShortArrayElements(env, p_halfword, 0);
   jshort         mode = j_mode;

   dbopen(base, pass, &mode, a_halfword);

   // for DBOPEN, all errors treated as exceptions
   if (a_halfword[0] != 0){
      ThrowException(env, "DBOPEN", a_halfword);
      return;
   }

   (*env)->ReleaseByteArrayElements(env, j_basename, base, 0);
   (*env)->ReleaseStringUTFChars(env, j_password, pass);
   (*env)->ReleaseShortArrayElements(env, p_halfword, a_halfword, 0);
}


/* DBCLOSE */
JNIEXPORT void JNICALL 
Java_com_hp_turboimage_DatabaseImpl_DBCLOSE (JNIEnv    *env, 
                                         jobject    this, 
                                         jbyteArray j_basename, 
                                         jstring    j_dataset, 
                                         jint       j_mode, 
                                         jobject    j_status) {

    // Get JNI parms into C types
   jbyte         *base = (*env)->GetByteArrayElements(env, j_basename, 0);
   const char *dataset = (*env)->GetStringUTFChars(env, j_dataset, 0);
   jclass   c_DBStatus = (*env)->GetObjectClass(env, j_status);
   jfieldID f_halfword = (*env)->GetFieldID(env, c_DBStatus, "halfword", "[S");
   jobject  p_halfword = (*env)->GetObjectField(env, j_status, f_halfword);
   jshort  *a_halfword = (*env)->GetShortArrayElements(env, p_halfword, 0);
   jshort         mode = j_mode;

    dbclose(base, dataset, &mode, a_halfword);

    if(a_halfword[0] < 0) {
       ThrowException(env, "DBCLOSE", a_halfword);
    }

   (*env)->ReleaseStringUTFChars(env, j_dataset, dataset); 
   (*env)->ReleaseByteArrayElements(env, j_basename, base, 0);
   (*env)->ReleaseShortArrayElements(env, p_halfword, a_halfword, 0);
}

/* DBLOCK */
JNIEXPORT void JNICALL 
Java_com_hp_turboimage_DatabaseImpl_DBLOCK (JNIEnv    *env, 
                                        jobject    this, 
                                        jbyteArray j_basename, 
                                        jobject    j_qualifier,
                                        jint       j_mode, 
                                        jobject    j_status) {

    // Get JNI parms into C types
    jbyte         *base = (*env)->GetByteArrayElements(env, j_basename, 0);
    jclass   c_DBStatus = (*env)->GetObjectClass(env, j_status);
    jfieldID f_halfword = (*env)->GetFieldID(env, c_DBStatus, "halfword", "[S");
    jobject  p_halfword = (*env)->GetObjectField(env, j_status, f_halfword);
    jshort  *a_halfword = (*env)->GetShortArrayElements(env, p_halfword, 0);
    short          mode = (short) j_mode;

    if ((mode == 3) || (mode == 4)) {
       // j_qualifier is a String (dataset lock)
       // qualif should be set to the string contents
       const char* qualif = (*env)->GetStringUTFChars(env, j_qualifier, 0);
       dblock(base, qualif, &mode, a_halfword);
       (*env)->ReleaseStringUTFChars(env, j_qualifier, qualif);
    } else {  // qualifier is a LockDescriptor object (item lock)
       // j_qualifier is a LockDescriptor object
       // qualif should be set to the contents of the buffer field
       jclass     ldclass = (*env)->GetObjectClass(env, j_qualifier);
       jfieldID      pbuf = (*env)->GetFieldID(env, ldclass, "buffer", "[B");
       jobject       abuf = (*env)->GetObjectField(env, j_qualifier, pbuf);
       jbyte      *qualif = (*env)->GetByteArrayElements(env, abuf, 0);
       dblock (base, qualif, &mode, a_halfword);
       (*env)->ReleaseByteArrayElements(env, abuf , qualif, 0);
    }

    if(a_halfword[0] < 0) {
       ThrowException(env, "DBLOCK", a_halfword);
    }

   (*env)->ReleaseByteArrayElements(env, j_basename, base, 0);
   (*env)->ReleaseShortArrayElements(env, p_halfword, a_halfword, 0);
}

/* DBUNLOCK */
JNIEXPORT void JNICALL 
Java_com_hp_turboimage_DatabaseImpl_DBUNLOCK (JNIEnv    *env, 
                                          jobject    this, 
                                          jbyteArray j_basename, 
                                          jstring    j_locks,
                                          jint       j_mode, 
                                          jobject    j_status) {

    // Get JNI parms into C types
    jbyte         *base = (*env)->GetByteArrayElements(env, j_basename, 0);
    const char *lockdes = (*env)->GetStringUTFChars(env, j_locks, 0);
    jclass   c_DBStatus = (*env)->GetObjectClass(env, j_status);
    jfieldID f_halfword = (*env)->GetFieldID(env, c_DBStatus, "halfword", "[S");
    jobject  p_halfword = (*env)->GetObjectField(env, j_status, f_halfword);
    jshort  *a_halfword = (*env)->GetShortArrayElements(env, p_halfword, 0);
    jshort         mode = j_mode;

    dbunlock(base, lockdes, &mode, a_halfword);

    if(a_halfword[0] < 0) {
       ThrowException(env, "DBUNLOCK", a_halfword);
    }


   (*env)->ReleaseStringUTFChars(env, j_locks, lockdes); 
   (*env)->ReleaseByteArrayElements(env, j_basename, base, 0);
   (*env)->ReleaseShortArrayElements(env, p_halfword, a_halfword, 0);

}

/* DBINFO */
JNIEXPORT void JNICALL 
Java_com_hp_turboimage_DatabaseImpl_DBINFO (JNIEnv * env, 
                                        jobject this, 
                                        jbyteArray j_basename, 
                                        jstring    j_qualifier, 
                                        jint       j_mode, 
                                        jobject    j_status, 
                                        jstring    j_buffer) {

// UNTESTED !!
// buffer should probably be replaced by an object - there are
// multiple formats, probably suggesting a class hierarchy. 
    // Get JNI parms into C types
    jbyte         *base = (*env)->GetByteArrayElements(env, j_basename, 0);
    const char *qualifier=(*env)->GetStringUTFChars(env, j_qualifier, 0);
    jclass   c_DBStatus = (*env)->GetObjectClass(env, j_status);
    jfieldID f_halfword = (*env)->GetFieldID(env, c_DBStatus, "halfword", "[S");
    jobject  p_halfword = (*env)->GetObjectField(env, j_status, f_halfword);
    jshort  *a_halfword = (*env)->GetShortArrayElements(env, p_halfword, 0);
    const char  *buffer = (*env)->GetStringUTFChars(env, j_buffer, 0);
    jshort         mode = j_mode;

    dbinfo(base, qualifier, &mode, a_halfword, (char *) buffer);

    if(a_halfword[0] < 0) {
       ThrowException(env, "DBINFO", a_halfword);
    }

   (*env)->ReleaseStringUTFChars(env, j_qualifier, qualifier); 
   (*env)->ReleaseStringUTFChars(env, j_buffer, buffer);
   (*env)->ReleaseByteArrayElements(env, j_basename, base, 0);
   (*env)->ReleaseShortArrayElements(env, p_halfword, a_halfword, 0);

}

