//: DatabaseImpl.java

package com.hp.turboimage;
 
// Following imports used only in main (testing) method
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.DataInputStream;  // temporary

/** This class is used to represent a TurboIMAGE database.  
 *  Methods are provided for opening, closing, and locking
 *  the database.  Most other functionality is provided by
 *  the Dataset object.
 *
 * @version 2.0 (RMI enabled) January 1998
 * @author  Mike Yawn
 */
public class DatabaseImpl extends java.rmi.server.UnicastRemoteObject
                       implements Database, Cloneable {

   // Intrinsic Parameters
   private byte[]   basename = new byte[26];
   private String   password = new String(";");
   private String   dummy    = new String(" ");  // an ignored parameter
   private DBStatus status = new DBStatus();

   static {
      // Note actual library name is libTurboIMAGE[_g].sl
      System.loadLibrary("TurboIMAGE");
   }

   /** Constructor to create database object, given a
    *  local database root file name.  If the first two
    *  characters you pass are not blanks, this routine
    *  will add blanks to the beginning of the name.  If
    *  the last character is not a semicolon, this routine
    *  will add one.  Filename should be in MPE syntax.
    *  (This routine should also accept HFS names, but doesn't yet)
    *
    * @param dbRootFile The name of the database root file.
    */
   public DatabaseImpl(String dbRootFile) throws java.rmi.RemoteException {
      setName(dbRootFile);
   }

   public void setName(String dbRootFile) {
      int endchar = dbRootFile.length();
      if (dbRootFile.startsWith("  ")){
         // 1.0 dbRootFile.getBytes(1, endchar, basename, 1);
         basename = dbRootFile.getBytes();
      } else {
         basename[0] = (byte)' '; basename[1] = (byte)' ';
         // 1.0 dbRootFile.getBytes(1, endchar, basename, 3);
         for (int i=0; i<endchar; i++) {
            basename[i+2] = (byte) dbRootFile.charAt(i);
         }
         //basename = ' ' + ' ' + dbRootFile.getBytes();
         endchar += 2;
      }
      if((basename[endchar-1] != ';') && (basename[endchar-1] != ' ')) {
         basename[endchar]= (byte)';'; 
      }
   }

   public byte[] getName() { return basename; }

   /** Set the database password.  Will add trailing semicolon
     * if not specified by the user.  If this method is not
     * called prior to an open call, the creator password will
     * be used as a default 
     * @param pass  The database password to use in subsequent 
     *              open calls.
     */
   public void setPassword(String pass){
      if(pass.endsWith(";")) {
         password = new String(pass);
      } else {
        password = new String(pass + ";");
      }
   }

   public DBStatus getStatus() {
      return status;
   }

   /** Open the database with an explicit mode.
    *  Numeric mode parameters will be more familiar to long-time
    *  users of the TurboIMAGE intrinsic calls.  Additional calls
    *  are available that use mnemonics to describe the open mode,
    *  which is more self-documenting.
    * @param mode   numeric value for open mode
    * @exception IMAGEException A TurboIMAGE error occured
    */
   public void open(int mode) throws IMAGEException {
      DBOPEN(basename, password, mode, status);
   }

   /** Open for modify, allow other readers and modifiers,
     *  with enforced locking 
     * @exception IMAGEException  A TurboIMAGE error occured
     */
   public void openSharedModify() throws IMAGEException {
      DBOPEN(basename, password, 1, status);
   }

   /** Open for update with other updaters allowed 
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public void openSemiExclusiveUpdate() throws IMAGEException {
      DBOPEN(basename, password, 2, status);
   }

   /** Open for update, allow readers 
     * @exception IMAGEException  A TurboIMAGE error occured */
   public void openSemiExclusiveUpdateAllowRead() throws IMAGEException {
      DBOPEN(basename, password, 2, status);  //(w/mode 6)
   }

   /** Open for modify; no other access 
     * @exception IMAGEException A TurboIMAGE error occured*/
   public void openExclusiveModify() throws IMAGEException {
      DBOPEN(basename, password, 3, status);
   }

   /** Open for modify, allow readers 
     * @exception IMAGEException A TurboIMAGE error occured*/
   public void openSemiExclusiveModifyAllowRead() throws IMAGEException {
      DBOPEN(basename, password, 4, status);
   }

   /** Open for read, allow modifiers 
     * @exception IMAGEException A TurboIMAGE error occured */
   public void openSharedRead() throws IMAGEException {
      DBOPEN(basename, password, 5, status);
   }

   /** Mode 6 opens are beyond my limited comprehension.  If
     * you see a need for it, this is it. 
     * @exception IMAGEException A TurboIMAGE error occured */
   public void openSemiExclusiveRead() throws IMAGEException {
      DBOPEN(basename, password, 6, status); //(w/mode 6)
   }

   /** Another way to open a database mode 6 
     * @exception IMAGEException A TurboIMAGE error occured*/
   public void openSemiExclusiveReadAllowUpdate() throws IMAGEException {
      DBOPEN(basename, password, 6, status); //(w/mode 2)
   }

   /** Another way to open a database mode 6 
     * @exception IMAGEException A TurboIMAGE error occured */
   public void openSemiExclusiveReadAllowModify() throws IMAGEException {
      DBOPEN(basename, password, 6, status); //(w/mode 4)
   }

   /** We really need another way to open a database mode 6 
     * @exception IMAGEException A TurboIMAGE error occured */
   public void openSemiExclusiveReadAllowReadOnly() throws IMAGEException {
      DBOPEN(basename, password, 6, status); //(w/mode 8)
   }

   /** Open the database for exclusive, read-only access 
     * @exception IMAGEException  A TurboIMAGE error occured */
   public void openExclusiveRead() throws IMAGEException {
      DBOPEN(basename, password, 7, status);
   }

   /** Open the database for read, allow other readers 
     * @exception IMAGEException  A TurboIMAGE error occured */
   public void openSemiExclusiveReadOnly() throws IMAGEException {
      DBOPEN(basename, password, 8, status);
   }

   /** Close the database 
     * @exception IMAGEException  A TurboIMAGE error occured */
   public void close() throws IMAGEException {
      DBCLOSE(basename, dummy, 1, status);
   }

   /** Lock the entire database.  Wait if the
     * lock is not immediately available.  
     * @exception IMAGEException  A TurboIMAGE error occured */
   public boolean lock() throws IMAGEException {
      DBLOCK(basename, dummy, 1, status);
      return (status.getStatus() == 0);
   }

   /** Attempt to lock the entire database.  If
     * the lock is not immediately available, returns false 
     * @exception IMAGEException  A TurboIMAGE error occured */
   public boolean attemptLock() throws IMAGEException {
      DBLOCK(basename, dummy, 2, status);
      return (status.getStatus() == 0);
   }

   /** Unlock the data base 
     * @exception IMAGEException A TurboIMAGE error occured */
   public boolean unlock() throws IMAGEException {
      DBUNLOCK(basename, dummy, 1, status);
      return (status.getStatus() == 0);
   }

   // Stubs that call TurboIMAGE Intrinsics
   public static native void DBOPEN(byte base[], String pass, int mode,
                              DBStatus status);
   public static native void DBCLOSE(byte base[], String dataset, int mode,
                              DBStatus status);
   public static native void DBLOCK(byte base[], Object qualifier, int mode,
                              DBStatus status);
   public static native void DBUNLOCK(byte base[], String qualifier, int mode,
                              DBStatus status);
   public static native void DBINFO(byte base[], String qualifier, int mode,
                              DBStatus status, String buffer);

   private static void test() {
      Database testdb;
      String   testname;
      String   testpass;
      Byte     modein;
      Integer  testmode;

      BufferedReader stdin = new BufferedReader(
                               new InputStreamReader(System.in));

      try {
         System.out.print("Enter database name :");
         testname = stdin.readLine();
         System.out.print("Enter password      :");
         testpass = stdin.readLine();
         System.out.print("Enter mode          :");
         testmode = new Integer(stdin.readLine());
         // above may throw NumberFormatException
         testdb   = new DatabaseImpl(testname);
         testdb.setPassword(testpass);
         testdb.open(testmode.intValue());
         System.out.println("now open: " + testname);
         // Verify our DBStatus array got back to us intact
         System.out.println(
                "User class: " + testdb.getStatus().getUserClass() );
         System.out.println(
                "DBG Size  : " + testdb.getStatus().getDBGSize() );
         System.out.println(
                "DBU Size  : " + testdb.getStatus().getDBUSize() );
         testdb.close();
      } catch (Exception e) {
         System.out.println("in main: " + e);
      }
   }

   private static void usage() {
      System.out.println(
        "Usage: java com.hp.turboimage.DatabaseImpl [options]");
      System.out.println();
      System.out.println("where options include: ");
      System.out.println("   -test             perform standalone testing");
      System.exit(0);
   } 

   public static void main(String[] argv) {
      boolean doTest = false;

      if (argv.length == 0) usage();

      for (int i=0; i<argv.length; i++) {
         if (argv[i].equals("-test"))     { doTest = true;        continue; }
         usage();
      }

      if (doTest) test();
      System.exit(0);  // not sure why we don't terminate otherwise
     
   }

}

