//: DatasetImpl.java

package com.hp.turboimage;

import java.net.*;
import java.io.*;
import java.rmi.RemoteException;

/** Class to represent TurboIMAGE datasets.  Provides object
 *  wrappers for DBFIND, DBGET, DBPUT, DBUPDATE, and DBDELETE;
 *  plus DBLOCK and DBUNLOCK at set or item level.<p>
 *
 *  The class works closely with the Entry class, which 
 *  represents the data buffer passed to put and update calls
 *  and received back from read calls.<p>
 *
 * @version 2.1  (RMI enabled) March 1998
 * @author  Mike Yawn
 */
public class DatasetImpl extends java.rmi.server.UnicastRemoteObject
                      implements Dataset, Cloneable {

   private byte[]   basename=new byte[26]; 
   private Database theDB;
   private String   setname;
   private String   dummy = new String();  // an ignored parameter
   private DBStatus status = new DBStatus();
   private int      requestedRecnum;

   /** Constructor to create dataset object 
     * @param db      The Database containing this set
     * @param setName String containining the IMAGE data set name
     */
   public DatasetImpl(Database db, String setName) 
                      throws java.rmi.RemoteException {
      theDB = db;
      try {
         basename = db.getName();
      } catch (java.rmi.RemoteException e) {
         e.printStackTrace();
         System.exit(0);
      }
      setname  = setName;
      if (! setname.endsWith(";"))
         setname = setname + ";";
   }

   /** Return the name of the data set represented by this class */
   public String getSetname() { return setname; }

   /** Re-read the current record (DBGET Mode 1) 
     * @param entry  Record contents will be returned in this object
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public boolean readCurrent(Entry entry) throws IMAGEException {
      try {
         byte[] buffer = new byte[entry.getBufferSize()];
         DBGET(basename, setname, 1, status, entry.getList(),
            buffer, dummy);
         if (status.getStatus() == 0) {
            entry.setBuffer(buffer);
            entry.enablePrevList();
            return true;
         }
         return false;
      } catch (RemoteException e) {
         System.out.println(e);
         return false;
      }
   };
 
   /** Read a dataset serially (DBGET Mode 2)
     * @param entry  Record contents will be returned in this object
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public boolean readSerial(Entry entry) throws IMAGEException {
      try {
         byte[] buffer = new byte[entry.getBufferSize()];
         DBGET(basename, setname, 2, status, entry.getList(),
            buffer, dummy);
         if (status.getStatus() == 0) {
            entry.setBuffer(buffer);
            entry.enablePrevList();
            return true;
         }
         return false;
      } catch (RemoteException e) {
         System.out.println(e);
         return false;
      }
   }; 

   /** Read a dataset serially in reverse order (DBGET Mode 3) 
     * @param entry  Record contents will be returned in this object
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public boolean readSerialBackward(Entry entry) throws IMAGEException {
      try {
         byte[] buffer = new byte[entry.getBufferSize()];
         DBGET(basename, setname, 3, status, entry.getList(),
            buffer, dummy);
         if (status.getStatus() == 0) {
            entry.setBuffer(buffer);
            entry.enablePrevList();
            return true;
         }
         return false;
      } catch (RemoteException e) {
         System.out.println(e);
         return false;
      } 
   };
 
   /** Read a dataset by record number (DBGET Mode 4)
     * @param entry  Record contents will be returned in this object
     * @param recnum Integer containing requested record number
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public boolean readDirected(Entry entry, int recnum) throws IMAGEException {
      // We pass a dummy arg to the DBGET routine, which will access
      // the following variable to get the actual arg value
      try { 
         byte[] buffer = new byte[entry.getBufferSize()];
         DBGET(basename, setname, 4, status, entry.getList(),
            buffer, new Integer(recnum));
         if (status.getStatus() == 0) {
            entry.setBuffer(buffer);
            entry.enablePrevList();
            return true;
         }
         return false;
      } catch (RemoteException e) {
         System.out.println(e);
         return false;
      }
   };

   /** Read the next entry on the current chain (DBGET Mode 5)
     * @param entry  Record contents will be returned in this object
     * @see   Dataset#find
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public boolean readChained(Entry entry) throws IMAGEException {
      try {
         byte[] buffer = new byte[entry.getBufferSize()];
         DBGET(basename, setname, 5, status, entry.getList(),
            buffer, dummy);
         if (status.getStatus() == 0) {
            entry.setBuffer(buffer);
            entry.enablePrevList();
            return true;
         }
         return false;
      } catch (RemoteException e) {
         System.out.println(e);
         return false;
      }
   }; 

   /** Read the previous entry on the current chain (DBGET Mode 6)
     * @param entry  Record contents will be returned in this object
     * @see   Dataset#find
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public boolean readChainedBackward(Entry entry) throws IMAGEException {
      try {
         byte[] buffer = new byte[entry.getBufferSize()];
         DBGET(basename, setname, 6, status, entry.getList(),
            buffer, dummy);
         if (status.getStatus() == 0) {
            entry.setBuffer(buffer);
            entry.enablePrevList();
            return true;
         }
         return false;
      } catch (RemoteException e) {
         System.out.println(e);
         return false;
      }
   };

   /** Read the entry having the specified key value (DBGET Mode 7)
     * @param entry  Record contents will be returned in this object
     * @param value  The requested value of the search item 
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public boolean readCalculated(Entry entry, Object value) 
                                 throws IMAGEException {
      try {
         byte[] buffer = new byte[entry.getBufferSize()];
         DBGET(basename, setname, 7, status, entry.getList(),
            buffer, value);
         if (status.getStatus() == 0) {
            entry.setBuffer(buffer);
            entry.enablePrevList();
            return true;
         }
         return false;
      } catch (RemoteException e) {
         System.out.println(e);
         return false;
      }
   };

   /** Read the entry occupying the primary address of a synonym
     * chain using the key item value specified.
     * @param entry  Record contents will be returned in this object
     * @param value  The requested value of the search item
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public boolean readPrimary(Entry entry, Object value) 
                              throws IMAGEException {
      try {
         byte[] buffer = new byte[entry.getBufferSize()];
         DBGET(basename, setname, 8, status, entry.getList(),
            buffer, value);
         if (status.getStatus() == 0) {
            entry.setBuffer(buffer);
            entry.enablePrevList();
            return true;
         }
         return false;
      } catch (RemoteException e) {
         System.out.println(e);
         return false;
      }
   };  

   /** Find all entries containing the requested search item value.
     * The chained read methods can be used to traverse the set
     * of entries found.
     * @param searchItem  The name of the field to be used as a key
     * @param value       Find entries with this value in the searchitem
     * @see   Dataset#readChained
     * @see   Dataset#readChainedBackward
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public boolean find(String searchItem, Object value)
                       throws IMAGEException {
      if (! searchItem.endsWith(";"))
         searchItem = searchItem + ";";

      // Strings must be padded to full length of item
      if (value instanceof String) {
         ItemInfo ii = new ItemInfo();
         ii.getInfo(theDB, searchItem, status);
         int datalen = ii.getSubitemLength();
         StringBuffer temp = new StringBuffer(datalen);
         for (int i=0; i<datalen; i++) {
            if (i < ((String)value).length())
               temp.append(((String)value).charAt(i));
            else
               temp.append(' ');
         }
         value = new String(temp);
      } 

      DBFIND(basename, setname, (short) 1, status, searchItem, value);
      return (status.getStatus() == 0);
   };

   /** Add an entry to the dataset            
     * @param entry  Contains the entry to be added
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public boolean put(Entry entry) throws IMAGEException {
      try {
         entry.toBuffer();
         DBPUT(basename, setname, 1, status, entry.getList(),
            entry.getBuffer());
         entry.enablePrevList();
      } catch (RemoteException e) {
         System.out.println(e);
         return false;
      }
      return (status.getStatus() == 0);
   };

   /** Update a previously retrieved dataset entry
     * @param entry  The entry to be updated 
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public boolean update(Entry entry) throws IMAGEException {
      try {
         entry.toBuffer();
         DBUPDATE(basename, setname, 1, status, entry.getList(),
               entry.getBuffer());
         entry.enablePrevList();
      } catch (RemoteException e) {
         System.out.println(e);
         return false;
      }
      return (status.getStatus() == 0);
   };

   /** Delete the current entry from the dataset 
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public boolean delete() throws IMAGEException {
      DBDELETE(basename, setname, 1, status);
      return (status.getStatus() == 0);
   };

   /** Lock the dataset.  If the set cannot be locked immediately,
     *  wait for it.  
     * @exception IMAGEException  A TurboIMAGE error occured
     */
   public boolean lock() throws IMAGEException {
      DBLOCK(basename, (Object) setname, 3, status);
      return (status.getStatus() == 0);
   };

   /** Lock the dataset.  If the set cannot be locked immediately,
     * the method returns 'false'. 
     * @exception IMAGEException  A TurboIMAGE error occured
     */
   public boolean attemptLock() throws IMAGEException {
      DBLOCK(basename, (Object) setname, 4, status);
      return (status.getStatus() == 0);
   };

   /** Lock the data entries described in the descriptor parameter.
     * Wait until the locks can be obtained.
     * @param descriptor  Describes the locks to be obtained
     * @exception IMAGEException A TurboIMAGE error occured
     */  
   public boolean lock(LockDescriptor descriptor) throws IMAGEException {
      DBLOCK(basename, (Object) descriptor, 5, status);
      return (status.getStatus() == 0); 
   };

   /** Lock the data entries described in the descriptor parameter.
     * If the locks cannot be obtained immediately, the method
     * returns 'false'.
     * @param descriptor  Describes the locks to be obtained
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public boolean attemptLock(LockDescriptor descriptor) 
                  throws IMAGEException {
      DBLOCK(basename, (Object) descriptor, 6, status);
      return (status.getStatus() == 0);
   };

   /** Unlock the dataset 
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public boolean unlock() throws IMAGEException {
      DBUNLOCK(basename, setname, 1, status);
      return (status.getStatus() == 0);
   };

   /** Close the dataset
     * @exception IMAGEException A TurboIMAGE error occurred
     */
   public boolean close() throws IMAGEException {
      DBCLOSE(basename, setname, 2, status);
      return (status.getStatus() == 0);
   };

   /** Reinitialize the dataset
     * @exception IMAGEException A TurboIMAGE error occurred
     */
   public boolean reinitialize() throws IMAGEException {
      DBCLOSE(basename, setname, 3, status);
      return (status.getStatus() == 0);
   }

   public static native void DBFIND(byte[] base, String set, int mode,
                              DBStatus status, String item, Object argument);
   public static native void DBGET(byte[] base, String set, int mode, 
                              DBStatus status, String list, 
                              byte[] buffer, Object arg);
   public static native void DBLOCK(byte[] base, Object qualifier, 
                              int mode, DBStatus status);
   public static native void DBPUT(byte[] base, String set, int mode,
                              DBStatus status, String list, byte[] buffer);
   public static native void DBUNLOCK(byte[] base, String qualifier, 
                              int mode, DBStatus status);
   public static native void DBUPDATE(byte[] base, String set, int mode,
                              DBStatus status, String list, byte[] buffer);
   public static native void DBDELETE(byte[] base, String qualifier, 
                              int mode, DBStatus status);
   public static native void DBCLOSE(byte[] base, String dataset,
                              int mode, DBStatus status);
}
