//: Entry.java

package com.hp.turboimage;

import java.util.Hashtable;
import java.util.Enumeration;
import java.rmi.*;

/** Class interface for TurboIMAGE entries
 *  Items of class Entry are returned by DBGET,
 *  and passed to DBPUT and DBUPDATE.  This
 *  class may be subclassed to create an
 *  entry for a particular dataset/list
 *  combination.<p>
 *
 * @version 2.2    July 1998
 * @author         Mike Yawn
 */
public interface Entry extends Remote {

  
  public void setDatabase(Database base) throws RemoteException;
  public void setDataset(Dataset set) throws RemoteException;
  public void enablePrevList() throws RemoteException;
  public void disablePrevList() throws RemoteException;

  /** addElement is used to add an element to the 'list'
    *  parameter for a dataset, and also to create an object
    *  to hold the value of the named item.  This variant
    *  is passed a name, type and length of an element.<p>
    *
    * @param name  String containing IMAGE item name
    * @param type  IMAGE data type.  Constants are defined as
    * @param occurs  The number of subitems in this element
    *  part of this class' definition.
    * @param length The length of this item in the data buffer
    */
   public void addElement(String name, int occurs, int type, int length)
                         throws RemoteException;

   /** This version of addElement handles non-compound
    *  items; name, type, and length are specified, and
    *  occurs is assumed to be one.
    * @param name  String containing IMAGE item name
    * @param type  IMAGE data type.  Constants are defined as
    * @param length The length of this item in the data buffer
    */
   public void addElement(String name, int type, int length) 
                          throws RemoteException;
   
   /** This version of addElement takes only the
    *  element name, and uses DBINFO to retrieve type
    *  information. 
    * @param name  String containing IMAGE item name
    */
   public void addElement(String name) throws RemoteException;

   /** allElements is used to build the "@;" list for a
    *  dataset.  It uses DBINFO to retrieve all items in
    *  the set, then calls addElement() for each item.
    */
   public void allElements() throws RemoteException;

   /** initElements should be called before any elements are
    *  added to the buffer definition by addElement calls.
    *  (Note: called for you when allElements method is used)
    */
   public void initElements() throws RemoteException;

   /** lockElements must be called after all elements have
    *  been added to the buffer using addElement.  (Note:
    *  called for you when AllElements method is used)
    */
   public void lockElements() throws RemoteException;

   /** Return a java.util.Enumeration with the names
     * of all Items in the entry.  These can each be
     * passed to method getValue() to return item values.
     * Ordering is random.
     */ 
   public Enumeration getElements() throws RemoteException;

   /** Before calling DBPUT or DBUPDATE, the Entry data type
    *  must be converted to a buffer (byte array) that will
    *  be used by the intrinsics.  This routine isn't called
    *  directly by the user, but rather by the put() and
    *  update() methods.  For internal use.
    * @exception IMAGEException  A TurboIMAGE error occured
    */
   public void toBuffer() throws IMAGEException, RemoteException;

   /** To prevent needless type conversions, we don't do a
     * 'fromBuffer' of an item until the user accesses it.
     * We keep track of this in the 'valid' field of each Item.
     * When we retrieve a new buffer, this routine is used to
     * mark all items in the buffer invalid (unconverted).  As
     * each field is accessed, a fromBuffer call will mark the
     * field as valid.  For internal use. */
   public void invalidateBuffer() throws RemoteException;

   /** After a DBGET call, the raw buffer returned by the
    *  must have its fields converted to appropriate Java types.
    *  This can be done on a field-by-field basis when a 
    *  getValue() method is called.  If it is known that all
    *  fields requested will be accessed, a performance
    *  improvement may result from doing the conversion all 
    *  at once, by calling this routine after the DBGET and
    *  before any calls to getValue().  If it is determined
    *  that there is little or no performance advantage to
    *  doing this, this method will be deprecated in future
    *  releases.
    * @exception IMAGEException A TurboIMAGE error occured
    */
   public void fromBuffer() throws IMAGEException, RemoteException;

   /** This fromBuffer variant is called by getItem(String), and
     * should not typically be called directly by the user.
     * @exception IMAGEException  A TurboIMAGE error occured
     */
   public void fromBuffer(String item) 
                            throws IMAGEException, RemoteException;

   /** This fromBuffer variant is called by getItem(String, int)
     * and should not typically be called directly by the user.
     * @exception IMAGEException  A TurboIMAGE error occured
     */
   public void fromBuffer(String item, int index)
                            throws IMAGEException, RemoteException;

//
// Methods to access instance variables
//

   /** This method returns the raw TurboIMAGE data buffer.
    *  End users should probably not be accessing it.
    */
   public byte[] getBuffer() throws RemoteException;

   /** This method returns the size of the data buffer */ 
   public int getBufferSize() throws RemoteException;

   /** This method sets the buffer contents */
   public void setBuffer(byte[] buffer) throws RemoteException;

   /** This method returns the TurboIMAGE item list.
    *  End users should probably not be accessing it.
    */
   public String getList() throws RemoteException;

   /** getValue is used to access the value of an item in
    *  an Entry.  A string containing the itemname is the key
    *  This variant is used for non-compound items.
    * @param key  The Item name
    * @exception IMAGEException A TurboIMAGE error occured
    */
   public Object getValue(String key) 
                  throws IMAGEException, RemoteException;

   /** getValue(String, int) is used to access the value of a
    *  compound item in an entry.  A String containing the
    *  itemname is the key, with the integer value specifying
    *  which occurence of the compound item.
    * @param key    The Item name
    * @param index  The subitem desired 
    * @exception IMAGEException A TurboIMAGE error occured
    */
   public Object getValue(String key, int index) 
                  throws IMAGEException, RemoteException;

   /** getItem is used to return an
     * Entry.Item object containing an 
     * item's value, plus its type, length, and other
     * attributes.  Intended for use by generic
     * query-type tools that don't have any pre-programmed
     * knowledge of the databases they access.
     * @param key  The Item name
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public Item getItem(String key) throws IMAGEException, RemoteException;

   /** getItemNames returns an array of Strings, with
     * the Strings containing the name of each Item in 
     * the Entry.   The entries will be in the same order
     * the data buffer was constructed, as opposed to the
     * random ordering of the getElements() enumeration.
     * This is preferable for Query-type tools wanting
     * to iterate through the fields in a predictable order.
     * @see Entry#getElements()
     */
   public String[] getItemNames() throws RemoteException;
 
   /** setValue is used to modify the value of an Item in
    *  an Entry. 
    *  This variant is used for non-compound items. 
    * @param key  A string containing the item name
    * @param value  An object providing the item value
    * @exception IMAGEException A TurboIMAGE error occured
    */
   public void setValue(String key, Object value) 
                  throws IMAGEException, RemoteException;

   /** This setValue() variant is used to modify the value of a
    *  compound item in an Entry.
    * @param key    A string containing the item name
    * @parm  index  The subitem to set the value of
    * @param value  An object providing the item value
    * @exception IMAGEException A TurboIMAGE error occured
    */
   public void setValue(String key, int index, Object value)
               throws IMAGEException, RemoteException;
}
