//: EntryImpl.java

package com.hp.turboimage;

import java.util.Hashtable;
import java.util.Enumeration;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;

/** Class interface for TurboIMAGE entries
 *  Items of class Entry are returned by DBGET,
 *  and passed to DBPUT and DBUPDATE.  This
 *  class may be subclassed to create an
 *  entry for a particular dataset/list
 *  combination.<p>
 *
 * @version 2.0 (RMI Enabled) January 1998 
 * @author         Mike Yawn
 */
public class EntryImpl extends UnicastRemoteObject 
                    implements Cloneable, Entry {

   /** TurboIMAGE E (IEEE format real) data type */
   public final static int IMAGE_TYPE_E = 1;
   /** TurboIMAGE I (integer) data type */
   public final static int IMAGE_TYPE_I = 2;
   /** TurboIMAGE J (integer COBOL COMPUTATIONAL) data type */
   public final static int IMAGE_TYPE_J = 3;
   /** TurboIMAGE K (unsigned) data type */
   public final static int IMAGE_TYPE_K = 4;
   /** TurboIMAGE P (comp-3) data type */
   public final static int IMAGE_TYPE_P = 5;
   /** TurboIMAGE R (HP format real) data type */
   public final static int IMAGE_TYPE_R = 6;
   /** TurboIMAGE U (uppercase alpha) data type */
   public final static int IMAGE_TYPE_U = 7;
   /** TurboIMAGE X (char) data type */
   public final static int IMAGE_TYPE_X = 8;
   /** TurboIMAGE Z (display) data type */
   public final static int IMAGE_TYPE_Z = 9;

   /** Constant value to specify all allowed items in dataset */
   static final String ALLITEMS  = "@;";
   /** Constant value to specify same list used by previous call */
   static final String PREVLIST  = "*;";

   private Database        db;
   private Dataset         ds;
   private DBStatus        status = new DBStatus();
   private String          list;
   private String          fullList;
   private StringBuffer    workList;
   private byte[]          buffer; 
   private Item            id;
   private Hashtable       items;        // key is itemname
   private String[]        itemNames = new String[1023]; 
   private int             curPosition;
   private int             index;        // item counter
   private boolean         prevListOK = false;

   /** Constructor sets local values for database, dataset
     * needed for dbinfo calls that build buffer
     * @param base  The Database object representing this database
     * @param set   The Dataset object representing this dataset
     */
   public EntryImpl(Database base, Dataset set) throws RemoteException {
      db = base;
      ds = set;
   }

   /** Update the Database field associated with this Entry.  Useful
     * for transaction-based systems that may not maintain the same
     * connection to a database during the life on an Entry
     */
   public void setDatabase(Database base) {
      db = base;
   }

   /** Update the Dataset field associated with this Entry.  Useful
     * for transaction-based systems that may not maintain the same
     * connection to a database during the life of an Entry
     */
   public void setDataset(Dataset set) {
      ds = set;
   }

   /** Enable use of the previous list (@) */
   public void enablePrevList() { 
      prevListOK = true; 
      list = PREVLIST;
   }

   /** Disable use of the previous list (@) */
   public void disablePrevList() {
      prevListOK = false;
      list = fullList;
   }
  
   /** addElement is used to add an element to the 'list'
    *  parameter for a dataset, and also to create an object
    *  to hold the value of the named item.  This variant
    *  is passed a name, type and length of an element.<p>
    *
    * @param name  String containing IMAGE item name
    * @param type  IMAGE data type.  Constants are defined as
    * @param occurs  The number of subitems in this element
    *  part of this class' definition.
    * @param length The length of this item in the data buffer
    */
   public void addElement(String name, int occurs, int type, int length){
       // Add item name to list
       workList.append(name.trim() + ",");

       // Add item to Hashtable
       id = new Item();
       id.value    = new Object[occurs];
       id.type     = type;
       id.occurs   = occurs;
       id.length   = length;
       id.position = curPosition;
       id.valid    = new boolean[occurs];
       curPosition += (length * occurs);
       items.put(name.trim(), id);
       itemNames[index++] = name.trim();
   }

   /** This version of addElement handles non-compound
    *  items; name, type, and length are specified, and
    *  occurs is assumed to be one.
    * @param name  String containing IMAGE item name
    * @param type  IMAGE data type.  Constants are defined as
    * @param length The length of this item in the data buffer
    */
   public void addElement(String name, int type, int length) {
      addElement(name, 1, type, length);
   }
   
   /** This version of addElement takes only the
    *  element name, and uses DBINFO to retrieve type
    *  information. 
    * @param name  String containing IMAGE item name
    */
   public void addElement(String name){
      ItemInfo iinfo = new ItemInfo();
      String   itype;
      int      len;

      iinfo.getInfo(db, name, status);

      // Convert all length types into bytes
      itype = iinfo.getItemType();
      if (itype.equals("X ") || itype.equals("U ") || 
          itype.equals("Z ") ) {
         // Length is already in bytes
         len = iinfo.getSubitemLength();
      } else if (itype.equals("P ")) { 
         // Length is in nibbles
         len = iinfo.getSubitemLength() / 2;
      } else {
         // Type E, I, J, K, or R; length is in halfwords
         len = iinfo.getSubitemLength() * 2;
      }
      addElement(name, iinfo.getSubitemCount(), 
                 imageTypeConvert(iinfo.getItemType()), len);
   }

   /** allElements is used to build the "@;" list for a
    *  dataset.  It uses DBINFO to retrieve all items in
    *  the set, then calls addElement() for each item.
    */
   public void allElements() {
      ItemList ilist = new ItemList();
      ItemInfo iinfo = new ItemInfo();
      String   itype;
      int len;

      try {
         ilist.getInfo(db, ds.getSetname(), status);
      } catch (java.rmi.RemoteException e) {
        e.printStackTrace();
         System.exit(0);
      }

      initElements();
      for (int i=1; i<=ilist.getItemCount(); i++) {
         iinfo.getInfo(db, ilist.getItemNumber(i), status);
         // Convert length from halfwords to bytes
         itype = iinfo.getItemType();
         if (itype.equals("X ") || itype.equals("U ") || 
             itype.equals("Z ") ) {
            len = iinfo.getSubitemLength();
         } else if (itype.equals("P ")) { 
            // Length is in nibbles
            len = iinfo.getSubitemLength() / 2;
         } else {
            // Type E, I, J, K, or R; length is in halfwords
            len = iinfo.getSubitemLength() * 2;
         }
         addElement(iinfo.getItemName(), iinfo.getSubitemCount(),
                    imageTypeConvert(iinfo.getItemType()), len );
      }
      lockElements();
      list = ALLITEMS; /* overwrite detailed list with shorthand */
   }

   /** Converts a two-byte string such as "X " to a 
     * corresponding integer value from this class' 
     * IMAGE_TYPE constants.  For internal use.
     * @param s  A two-byte string (typically returned by 
     *           DBINFO) describing the item type
     */
   int imageTypeConvert(String s) {
      if (s.equals("E ")) return IMAGE_TYPE_E;
      if (s.equals("I ")) return IMAGE_TYPE_I;
      if (s.equals("J ")) return IMAGE_TYPE_J;
      if (s.equals("K ")) return IMAGE_TYPE_K; 
      if (s.equals("R ")) return IMAGE_TYPE_R;
      if (s.equals("P ")) return IMAGE_TYPE_P;
      if (s.equals("U ")) return IMAGE_TYPE_U; 
      if (s.equals("X ")) return IMAGE_TYPE_X; 
      if (s.equals("Z ")) return IMAGE_TYPE_Z;    

      System.out.println("imageTypeConvert implementation error: " 
         + "Type = " + s);
      return 0;
   }

   /** initElements should be called before any elements are
    *  added to the buffer definition by addElement calls.
    *  (Note: called for you when allElements method is used)
    */
   public void initElements(){
      workList = new StringBuffer();
      items = new Hashtable(1023);
      curPosition = 0;
      index = 0;
   }

   /** lockElements must be called after all elements have
    *  been added to the buffer using addElement.  (Note:
    *  called for you when allElements method is used)
    */
   public void lockElements(){
      buffer = new byte[curPosition];
      workList.setCharAt(workList.length()-1,';');
      fullList = workList.toString();
      disablePrevList();
   }

   /** Return a java.util.Enumeration with the names
     * of all Items in the entry.  These can each be
     * passed to method getValue() to return item values.
     * Ordering is random.
     */ 
   public Enumeration getElements() {
      return items.keys();
   }

//
// TurboIMAGE buffer to/from Java Object conversion
//

   /** Before calling DBPUT or DBUPDATE, the Entry data type
    *  must be converted to a buffer (byte array) that will
    *  be used by the intrinsics.  This routine isn't called
    *  directly by the user, but rather by the put() and
    *  update() methods.  For internal use.
    * @exception IMAGEException A TurboIMAGE error occured
    */
   public void toBuffer() throws IMAGEException {
      Enumeration  e    = items.keys();
      String       name = new String();
      Item         info = new Item();
      byte[]       itemBytes;
      int start, len;

      while (e.hasMoreElements()){
         name = (String) e.nextElement();
         info = (Item) items.get(name);
//System.out.println("Processing " + name);
//System.out.println("Length is " + info.length);
         for (int si=1; si<=info.getSubitemCount(); si++) {
            switch(info.type){
               case IMAGE_TYPE_E: // IEEE Real
                  if (info.length <= 4) {
                     // Source object is a float
                     itemBytes = TypeConverter.floatToBytes(
                                     ((Float) info.value[si-1]).floatValue(),
                                     TypeConverter.IEEE_FORMAT);
//System.out.println("converted to float (IEEE)");
                     break;
                  }
                  if (info.length <= 8) {
                     // Source object is a double
                     itemBytes = TypeConverter.doubleToBytes(
                                     ((Double) info.value[si-1]).doubleValue(),
                                     TypeConverter.IEEE_FORMAT);
//System.out.println("converted to double (IEEE)");
                     break;
                  }
                  throw new IMAGEException
                            ("Can't convert real value > 8 bytes");

               case IMAGE_TYPE_R:
                  if (info.length <= 4) {
                     // Source object is a float
                     itemBytes = TypeConverter.floatToBytes(
                                     ((Float) info.value[si-1]).floatValue(),
                                     TypeConverter.HP_FORMAT);
//System.out.println("converted to float (HP)");
                     break;
                  }
                  if (info.length <= 8) {
                     // Source object is a double
                     itemBytes = TypeConverter.doubleToBytes(
                                     ((Double) info.value[si-1]).doubleValue(),
                                     TypeConverter.HP_FORMAT);
//System.out.println("converted to double (HP)");
                     break;
                  }
                  throw new IMAGEException
                            ("Can't convert real value > 8 bytes");

               case IMAGE_TYPE_I: // integer
               case IMAGE_TYPE_J: // COBOL COMPUTATIONAL
               case IMAGE_TYPE_K: // FORTRAN LOGICAL
                  if (info.length <= 2) {
                     // Source object is a short
                     itemBytes = TypeConverter.shortToBytes(
                                     ((Short) info.value[si-1]).shortValue());
//System.out.println("converted to short");
                     break;
                  }
                  if (info.length <= 4) {
                     // Source object is an int
                     itemBytes = TypeConverter.intToBytes(
                                     ((Integer) info.value[si-1]).intValue());
//System.out.println("converted to integer");
                     break;
                  }
                  if (info.length <= 8) {
                     // Source object is a long
                     itemBytes = TypeConverter.longToBytes(
                                     ((Long) info.value[si-1]).longValue());
//System.out.println("converted to long");
                     break;
                  }
                  throw new IMAGEException
                     ("Can't convert integer value > 8 bytes");

               case IMAGE_TYPE_P: // Packed decimal (COBOL COMP-3)
                  itemBytes = TypeConverter.stringToPackedBytes(
                              (String) info.value[si-1], info.length);
//System.out.println("converted to " + info.length + " packed bytes");
                  break;

               case IMAGE_TYPE_U: // uppercase alpha
               case IMAGE_TYPE_X: // alpha
               case IMAGE_TYPE_Z: // display numeric
//System.out.println("converted to String");
                  itemBytes = ((String) info.value[si-1]).getBytes();
                  break;

               default:
                  throw new IMAGEException
                     ("Entry.toBuffer() doesn't understand type " + info.type);
            } // switch
//System.out.println("itemBytes contains " + new String(itemBytes));
//System.out.println("itemBytes length is " + itemBytes.length);
//System.out.println("subitem count is " + si);
//System.out.println("buffer length is " + buffer.length);
//System.out.println("item starts in buffer at " + info.position);
//System.out.println("copying " + info.length + " bytes to " + 
//                    (info.position + (info.length * (si-1))) );
            System.arraycopy(itemBytes, 0,  buffer, 
               info.position + (info.length * (si-1)), itemBytes.length );
            if (itemBytes.length < info.length) {
               // Space fill the remaining itemBytes
               for (int i=itemBytes.length; i<info.length; i++) {
                  buffer[info.position + (info.length * (si-1)) + i] 
                                  = (byte)' ';
               }
            } 
         } // for loop
      } // while loop
   } // method

   /** To prevent needless type conversions, we don't do a
     * 'fromBuffer' of an item until the user accesses it.
     * We keep track of this in the 'valid' field of each Item.
     * When we retrieve a new buffer, this routine is used to
     * mark all items in the buffer invalid (unconverted).  As
     * each field is accessed, a fromBuffer call will mark the
     * field as valid.  For internal use. */
   public void invalidateBuffer() {
      Enumeration e   = items.keys();
      String      name = new String();
      Item        info = new Item();

      while (e.hasMoreElements()) {
         name = (String) e.nextElement();
         info = (Item) items.get(name);
         // booleans are by default initialized to false
         info.valid = new boolean[info.occurs];
      }
   }

   /** After a DBGET call, the raw buffer returned by the
    *  must have its fields converted to appropriate Java types.
    *  This can be done on a field-by-field basis when a 
    *  getValue() method is called.  If it is known that all
    *  fields requested will be accessed, a performance
    *  improvement may result from doing the conversion all 
    *  at once, by calling this routine after the DBGET and
    *  before any calls to getValue().  If it is determined
    *  that there is little or no performance advantage to
    *  doing this, this method will be deprecated in future
    *  releases.
    * @exception IMAGEException A TurboIMAGE error occured
    */
   public void fromBuffer() throws IMAGEException {
      Enumeration e    = items.keys();
      String      name = new String();
      Item        info = new Item();

      while (e.hasMoreElements()){
         name = (String) e.nextElement();
         info = (Item) items.get(name);
         for(int si=0; si<info.getSubitemCount(); si++) {
            fromBuffer(name, si);
         }
      }
   }

   /** This fromBuffer method is used to convert a single
     * non-compound item from buffer (bytearray) format to
     * the correct Java type.  For internal use by the 
     * getValue(name) method.
     * @param name  The name of the Item to convert
     * @exception IMAGEException  A TurboIMAGE error occured
     */
   public void fromBuffer(String name) throws IMAGEException {
      fromBuffer(name, 0);
   }

   /** This fromBuffer method is used to convert a single
     * subitem of a compound item from buffer (bytearray)
     * format to the correct Java type.  For internal use
     * by the getValue(name, index) method.
     * @param name   The name of the Item to convert
     * @param index  Which subitem to convert
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public void fromBuffer(String name, int index)
                  throws IMAGEException  {
      // subitem count starts at one, but index is 0 based
      int si = index+1;  // SubItems count from 1; java indexes from 0
      Item info = new Item();
      int start, len; 

      info = (Item) items.get(name);
      byte[] itemBytes = new byte[info.length];
      System.arraycopy(buffer, (info.position + (info.length * (si-1))), 
                       itemBytes, 0, info.length);

      switch(info.type){
         case IMAGE_TYPE_E: // IEEE Real
            if (info.length <= 4) {
               // Target object will be a float
               info.value[si-1] = (Object)
                   new Float(TypeConverter.bytesToFloat(itemBytes, 
                                        TypeConverter.IEEE_FORMAT));
               break;
               }
            if (info.length <= 8) {
               // Target object will be a double
               info.value[si-1] = (Object)
                   new Double(TypeConverter.bytesToDouble(itemBytes,
                                         TypeConverter.IEEE_FORMAT));
               break;
            }
            throw new IMAGEException
               ("Can't convert real value > 8 bytes");
            
         case IMAGE_TYPE_R: // HP 3000 Real 
            if (info.length <= 4) {
               // Target object will be a float
               info.value[si-1] = (Object)
                   new Float(TypeConverter.bytesToFloat(itemBytes, 
                                        TypeConverter.HP_FORMAT));
               break;
               }
            if (info.length <= 8) {
               // Target object will be a double
               info.value[si-1] = (Object)
                   new Double(TypeConverter.bytesToDouble(itemBytes,
                                         TypeConverter.HP_FORMAT));
               break;
            }
            throw new IMAGEException
               ("Can't convert real value > 8 bytes");
            

         case IMAGE_TYPE_I:  // integer
         case IMAGE_TYPE_J:  // COBOL COMPUTATIONAL
         case IMAGE_TYPE_K:  // FORTRAN LOGICAL
            if (info.length <= 2) {
               // Target object will be a short
               info.value[si-1] = (Object)
                   new Short(TypeConverter.bytesToShort(itemBytes));
               break;
            } 
            if (info.length <= 4) {
               // Target object will be an int
               info.value[si-1] = (Object)
                   new Integer(TypeConverter.bytesToInt(itemBytes));
               break;
            }
            if (info.length <= 8) {
               // Target object will be a long
               info.value[si-1] = (Object)
                   new Long(TypeConverter.bytesToLong(itemBytes));
               break;
            }
            throw new IMAGEException
               ("Can't convert integer value > 8 bytes");

         case IMAGE_TYPE_P:  // Packed Decimal (comp-3) 
            info.value[si-1] = (Object)
                TypeConverter.packedBytesToString(itemBytes);
            break;

         case IMAGE_TYPE_U: // uppercase alpha
         case IMAGE_TYPE_X: // alpha
         case IMAGE_TYPE_Z: // display numeric
            // Target object will be a String
            info.value[si-1] = (Object) new String(itemBytes);
            break;
         default:
            throw new IMAGEException
               ("Entry.fromBuffer() doesn't understand type " 
                 + info.type);
       } // end switch
       info.valid[si-1] = true;
       items.put(name, info); // updates the value field
}

//
// Methods to access instance variables
//

   /** This method returns the raw TurboIMAGE data buffer.
    *  End users should probably not be accessing it.
    */
   public byte[] getBuffer() { return buffer; }

   /** This method returns the data buffer size */
   public int getBufferSize() { return buffer.length; }

   /** This method sets the buffer contents */
   public void setBuffer(byte[] newbuf) {
      buffer = newbuf;
      invalidateBuffer();
   }

   /** This method returns the TurboIMAGE item list
    *  End users should probably not be accessing it.
    */
   public String getList()   { return list;   }

   /** getValue is used to access the value of an item in
    *  an Entry.  A string containing the itemname is the key
    *  This variant is used for non-compound items.
    * @param key  The Item name
    * @exception IMAGEException A TurboIMAGE error occured
    */
   public Object getValue(String key) throws IMAGEException {
      Item id = new Item();
      id = (Item) items.get(key);
      if (id == null) throw new IMAGEException("No such item: " + key);
      if (! id.valid[0]) {                             
         fromBuffer(key);
      }
      return id.value[0];
   }

   /** getValue(String, int) is used to access the value of a
    *  compound item in an entry.  A String containing the
    *  itemname is the key, with the integer value specifying
    *  which occurence of the compound item.
    * @param key    The Item name
    * @param index  The subitem desired 
    * @exception IMAGEException  A TurboIMAGE error occured
    */
   public Object getValue(String key, int index) throws IMAGEException {
      Item id = new Item();
      id = (Item) items.get(key);
      if (id == null) throw new IMAGEException("No such item: " + key);
      if (! id.valid[index]) {
         fromBuffer(key, index);
      }
      return id.value[index];
   }

   /** getItem is used to return an
     * Entry.Item object containing an 
     * item's value, plus its type, length, and other
     * attributes.  Intended for use by generic
     * query-type tools that don't have any pre-programmed
     * knowledge of the databases they access.
     * @param key  The Item name
     * @exception IMAGEException A TurboIMAGE error occured
     */
   public Item getItem(String key) throws IMAGEException {
      Item id = new Item();
      id = (Item) items.get(key);
      if (id == null) throw new IMAGEException("No such item: " + key);
      return id;
   }

   /** getItemNames returns an array of Strings, with
     * the Strings containing the name of each Item in 
     * the Entry.   The entries will be in the same order
     * the data buffer was constructed, as opposed to the
     * random ordering of the getElements() enumeration.
     * This is preferable for Query-type tools wanting
     * to iterate through the fields in a predictable order.
     * @see Entry#getElements()
     */
   public String[] getItemNames() { return itemNames; }
 
   /** setValue is used to modify the value of an Item in
    *  an Entry. 
    *  This variant is used for non-compound items. 
    * @param key  A string containing the item name
    * @param value  An object providing the item value
    * @exception IMAGEException A TurboIMAGE error occured
    */
   public void setValue(String key, Object value) throws IMAGEException {
      //Item id = new Item();
      Item id = (Item) items.get(key);
      if (id == null) throw new IMAGEException("No such item: " + key);
      id.value[0] = value;
      items.put(key, id);
   }

   /** This setValue() variant is used to modify the value of a
    *  compound item in an Entry.
    * @param key    A string containing the item name
    * @param index  The subitem to set the value of
    * @param value  An object providing the item value
    * @exception IMAGEException  A TurboIMAGE error occured
    */
   public void setValue(String key, int index, Object value)
               throws IMAGEException {
      Item id = new Item();
      id = (Item) items.get(key);
      if (id == null) throw new IMAGEException("No such item: " + key);
      id.value[index] = value;
      items.put(key,id);
   }

}
