//: LockDescriptor.java

package com.hp.turboimage;

import java.io.*;

/** This class represents a lock descriptor array used to
  * perform item-level locking.
  *
  * @version 2.3  August 1998
  * @author  Mike Yawn
  * @author  Jim Fennell (InterPay, Inc. Comments to: JFennell@InterPay.com)
  */
public class LockDescriptor implements java.io.Serializable {

   private byte buffer[] = new byte[4094];  // TI maximum
   private short  numEntries;
   private int  bufferused;
   private static byte[] fill = new byte[80];

   /** Constructor for lockdescriptor object */
   public LockDescriptor() {
      setNumEntries((short)0);
      bufferused = 2;
      for (int i=0; i<16; i++) fill[i]=(byte)' ';
   }

   /** Free any unneeded space in the lockdescriptor
     * buffer.  This may improve performance; no testing
     * has confirmed this as yet.
     */
   public void truncate() {
      // chop the buffer down to size.  Currently just
      // testing whether this makes a difference.
      byte newbuf[] = new byte[bufferused];
      System.arraycopy(buffer, 0, newbuf, 0, bufferused);
      buffer = newbuf;
   }

   // Internal helper routine
   public int getItemLength(String itemName) {
      /* For X type items, it would be nice to loop up the
       * item length, rather than require it to be passed.
       * However, with the current design, we don't have the
       * database object that would be necessary to allow us
       * to call dbinfo.
       *
       * ItemItem ii = new ItemInfo();
       * ii.getInfo(db, itemName.toUpperCase(), new DBStatus());
       * return ii.getSubItemLength();
       */
      return -1;
   }

   /** A convenience for Dataset.DBLOCK() native method
     */
   byte[] getBuffer() { return buffer; }

   /** addDatasetLock is the method you need to call to create a lock
     * descriptor for a dataset (set level) lock.  It may be called multiple
     * times if more than one dataset lock is being specified.
     */
   public void addDatasetLock(String ds) {
      byte[] temp;  // to hold converted strings
      int itemstart, itemlen;
      int tofill;

      setNumEntries(++numEntries);
      itemstart = bufferused;

      // dataset in halfwords 2-9 (bytes 2-18)
      temp = ds.getBytes();
      System.arraycopy(temp, 0, buffer, bufferused + 2, temp.length);
      tofill = 16-temp.length;
      System.arraycopy(fill, 0, buffer, bufferused + 2 + temp.length, tofill);
      // Note; values below reserve room for length, added last
      bufferused += 18; itemlen = 18;

      // item is '@' for a dataset lock
      buffer[bufferused++] = (byte)'@';
      buffer[bufferused++] = (byte)' '; // preserves halfword boundaries
      itemlen = 20;

      // length in entry goes at entry start
      byte[] b = new byte[2];
      b = TypeConverter.shortToBytes((short)((itemlen+1)/2));
      System.arraycopy(b, 0, buffer, itemstart, 2);

      // number of entries goes at lockdescriptor start
      b = TypeConverter.shortToBytes(numEntries);
      System.arraycopy(b, 0, buffer, 0, 2);

   }

   /** addElement has been deprecated in favor of the more precise
     * terminology of addPredicateLock.
     * @deprecated
     */
   public void addElement(String ds, String item, String relop, Object value)
               throws IMAGEException {
      addPredicateLock(ds, item, relop, value);
   }

   /** addElement has been deprecated in favor of the more precise
     * terminology of addPredicateLock.
     * @deprecated
     */
   public void addElement(String ds, String item, String relop,
                          Object value, int valueLength)
               throws IMAGEException {
      addPredicateLock(ds, item, relop, value, valueLength);
   }

   /** addPredicateLock is the main method you need to call to create
     * a lock descriptor for a predicate (entry level) lock.  It needs
     * to be called at least once, and may be called multiple times if
     * more than one predicate is being specified.  This variant can be
     * used for all but X type fields; for X type fields you must also
     * specify the length of the field being locked.
     * @exception  IMAGEException A TurboIMAGE error occured, or an item
     *             length was required and not passed.
     */
   public void addPredicateLock(String ds, String item, String relop, 
                                Object value)
               throws IMAGEException {
      addPredicateLock(ds, item, relop, value, 0);
   }

   /** This addPredicateLock variant allows the length of the item field to
     * be specified; this is needed for X type fields to work.
     * @exception  IMAGEException A TurboIMAGE error occured, or an item
     *             length of zero was passed
     */
   public void addPredicateLock(String ds, String item, String relop,
                          Object value, int valueLength)
               throws IMAGEException {
      byte[] temp;  // to hold converted strings
      int itemstart, itemlen;
      int tofill;

      setNumEntries(++numEntries);
      itemstart = bufferused;

      // dataset in halfwords 2-9 (bytes 2-18)
      temp = ds.getBytes();
      System.arraycopy(temp, 0, buffer, bufferused + 2, temp.length);
      tofill = 16-temp.length;
      System.arraycopy(fill, 0, buffer, bufferused + 2 + temp.length, tofill);
      // Note; values below reserve room for length, added last
      bufferused += 18; itemlen = 18;

      // item in halfwords 10-17 (bytes 19-34)
      if (item != null) {
         temp = item.getBytes();
         System.arraycopy(temp, 0, buffer, bufferused, temp.length);
         bufferused += temp.length;
         tofill = 16-temp.length;
         System.arraycopy(fill, 0, buffer, bufferused, tofill);
         bufferused += tofill; itemlen = 34;

         // relop in halfword 18 (bytes 35-36)
         temp = relop.getBytes();
         System.arraycopy(temp, 0, buffer, bufferused, temp.length);
         if (temp.length == 1) buffer[bufferused+1] = (byte)' ';
         bufferused += 2; itemlen = 36;

         // value in halfword 19 - ? (bytes 37-?)
         if (value instanceof String) {
            temp = ((String)value).getBytes();
            System.arraycopy(temp, 0, buffer, bufferused, temp.length);
            if (valueLength == 0) {
               throw new IMAGEException("Value Length required for X items");
            }
            tofill = valueLength - temp.length;
            System.arraycopy(fill, 0, buffer, bufferused+temp.length, tofill);
            bufferused += valueLength; itemlen += valueLength;
         } else if (value instanceof Short) {
            temp = TypeConverter.shortToBytes(((Short)value).shortValue());
            System.arraycopy(temp, 0, buffer, bufferused, 2);
            bufferused += 2; itemlen = 38;
         } else if (value instanceof Integer) {
            temp = TypeConverter.intToBytes(((Integer)value).intValue());
            System.arraycopy(temp, 0, buffer, bufferused, 4);
            bufferused += 4; itemlen = 40;
         } else if (value instanceof Long) {
            temp = TypeConverter.longToBytes(((Long)value).longValue());
            System.arraycopy(temp, 0, buffer, bufferused, 8);
            bufferused += 8; itemlen = 44;
         } else if (value instanceof Float) {
            temp = TypeConverter.floatToBytes(((Float)value).floatValue(),
// NOTE: should either check data type, or give user another parm for format
                                               TypeConverter.IEEE_FORMAT);
            System.arraycopy(temp, 0, buffer, bufferused, 4);
            bufferused += 4; itemlen = 40;
         } else if (value instanceof Double) {
            temp = TypeConverter.doubleToBytes(((Double)value).doubleValue(),
// NOTE: should either check data type, or give user another parm for format
                                                 TypeConverter.IEEE_FORMAT);
            System.arraycopy(temp, 0, buffer, bufferused, 8);
            bufferused += 8; itemlen = 44;
         }
      }
      // length in entry goes at entry start
      byte[] b = new byte[2];
      b = TypeConverter.shortToBytes((short)((itemlen+1)/2));
      System.arraycopy(b, 0, buffer, itemstart, 2);

      // number of entries goes at lockdescriptor start
      b = TypeConverter.shortToBytes(numEntries);
      System.arraycopy(b, 0, buffer, 0, 2);

   }

   /** Used internally to keep track of how many predicates are
     * specified in this LockDescriptor */
   private void setNumEntries(short n) {
      numEntries = n;
   }

   /** Display the contents of the Lock Descriptor Array */
   public void display() {
      boolean atEnd = false;
      int     count = 0;
      int     position = 0;
      int     ldent = 0;
      int     entrylen = 0;
      byte[]  byte2 = new byte[2];
      byte[]  byte4 = new byte[4];
      byte[]  byte16 = new byte[16];
      byte[]  byte80 = new byte[80];

      // Display Array Length
      System.arraycopy(buffer, position, byte2, 0, 2);
      position += 2;
      ldent = TypeConverter.bytesToShort(byte2);
      System.out.println("Lock Descriptor entries: " + ldent);

      // Loop through Lock Descriptors
      while (count < numEntries) {
         System.out.println("Lock Descriptor #" + ++count);
         // Entry Length
         System.arraycopy(buffer, position, byte2, 0, 2);
         position += 2;
         entrylen = TypeConverter.bytesToShort(byte2);
         System.out.println("   Entry length: " + entrylen + " halfwords");
         // Dataset
         System.arraycopy(buffer, position, byte16, 0, 16);
         position += 16;
         System.out.println("   Dataset     : " + new String(byte16));
         if (buffer[position] == '@') {
            System.out.println("   DATASET LOCK");
            position += 2;
         } else {
            // Data Item
            System.arraycopy(buffer, position, byte16, 0, 16);
            position += 16;
            System.out.println("   Data item   : " + new String(byte16));
            // Relation
            System.arraycopy(buffer, position, byte2, 0, 2);
            position += 2;
            System.out.println("   Relation    : " + new String(byte2));
            // Value
// NOTE: this works only for X-type data
            int valuelen = 2*entrylen - 36;
            System.arraycopy(buffer, position, byte80, 0, valuelen);
            int tofill = 80-valuelen;
            System.arraycopy(fill, 0, byte80, valuelen, tofill);
            position += valuelen;
            System.out.println("   Value       : " + new String(byte80));
         }
      }
      System.out.println("------- END ---------");
   }

   /** Standalone test for this class */
   public static void main(String[] args) {

      BufferedReader stdin = new BufferedReader(
                             new InputStreamReader(System.in));
      String testname, testpass, testset, testitem, testvalue;
      Database db;
      Dataset ds;
      DBStatus status = new DBStatus();
      LockDescriptor locks = new LockDescriptor();
      int len = 0;

      try {
      System.out.print("Enter database name: ");
      testname = stdin.readLine();
      db = new DatabaseImpl(testname.toUpperCase());
      System.out.print("Enter password: ");
      testpass = stdin.readLine();
      db.setPassword(testpass);
      try {
         db.openSharedRead();
      } catch (IMAGEException e) {
         System.out.println("failed to open " + e);
      }
      System.out.print("Enter set name: ");
      testset = stdin.readLine() + ";";
      ds = new DatasetImpl(db, testset);
      System.out.print("Enter item name: ");
      testitem = stdin.readLine() + ";";
      System.out.print("Enter value: ");
      testvalue = stdin.readLine();
      System.out.print("Enter length of field: ");
      try {
         len = Integer.parseInt(stdin.readLine());
      } catch (NumberFormatException e) { e.printStackTrace(); }

      try {
         locks.addElement(testset, testitem, "=", testvalue, len);
         locks.truncate();
         locks.display(); // debugging only
         ds.lock(locks);
      } catch (IMAGEException e) { e.printStackTrace(); }

      System.out.print(
           "Check for lock using DBUTIL, return to unlock & continue");
      stdin.readLine();

      try {
         ds.unlock();
      } catch (IMAGEException e) { e.printStackTrace(); }
      } catch (Exception x) { System.out.println(x); }

      System.exit(0);
   }
}
