//: TypeConverter.java

package com.hp.turboimage;
import java.io.*;  

/** This class provides conversion routines that
  * can take an array of bytes and convert it to a
  * java datatype that most closely matches the
  * IMAGE type for that data, and vice versa.
  *
  * @version 2.1.1  April 1998
  * @author         Mike Yawn
  */
public class TypeConverter {

static {
   System.loadLibrary("TurboIMAGE");
}

   /** Specifies IEEE format for conversion of reals */
   public static final int IEEE_FORMAT = 1;
   /** Specifies HP (MPE/V) format for conversion of reals */
   public static final int HP_FORMAT   = 2;

   /** Return 2-byte byte array as a short */
   public native static short bytesToShort(byte[] in);

   /** Return 4-byte byte array as an int */
   public native static int bytesToInt(byte[] in);

   /** Return 8-byte byte array as a long */
   public native static long bytesToLong(byte[] in);

   /** Return 4-byte byte array as a float */
   public native static float bytesToFloat(byte[] in, int format);

   /** Return 8-byte byte array as a double */
   public native static double bytesToDouble(byte[] in, int format);

   /** Return short as a 2-byte byte array */
   public native static byte[] shortToBytes(short in);

   /** Return int as a 4-byte byte array */
   public native static byte[] intToBytes(int in);

   /** Return long as an 8-byte byte array */
   public native static byte[] longToBytes(long in);

   /** Return float as a 4-byte byte array */
   public native static byte[] floatToBytes(float in, int format);

   /** Return double as an 8-byte byte array */
   public native static byte[] doubleToBytes(double in, int format);

   /** Return String as a packed byte array */
   public native static byte[] stringToPackedBytes(String in, int len);
     
   /** Return array of packed bytes as a String */
   public native static String packedBytesToString(byte[] in);

   public static void main(String[] args) 
                      throws IOException {
      BufferedReader stdin = 
          new BufferedReader(new InputStreamReader(System.in));
      byte[] result;
      String status;
      int    function, format=0;

      for (;;) {
         System.out.println("Test type conversion for which type: ");
         System.out.println("   1. Short     (I, J) ");
         System.out.println("   2. Integer   (I2, J2)");  
         System.out.println("   3. Long      (I4, J4)");
         System.out.println("   4. Packed Decimal (P) ");
         System.out.println("   5. Float as IEEE Real     (E2) ");
         System.out.println("   6. Float as HP MPE/V Real (R2) ");
         System.out.println("   7. Double as IEEE Real    (E4) ");
         System.out.println("   8. Double as HP MPE/V Real (R4)");
         System.out.println(" ");
         System.out.println("   9. [ to exit ] ");
         function = Integer.parseInt(stdin.readLine());

         if (function == 9) System.exit(0);

         switch (function) {
            case 1:  // Short
               try {
                  System.out.print("Enter short value: ");
                  short s1 = Short.parseShort(stdin.readLine());
                  result = shortToBytes(s1);
                  short s2 = bytesToShort(result);
                  if (s1 == s2) status = " OK "; else status = " FAIL ";
                  System.out.println("Retrieved " + s2 + status);
               } catch (NumberFormatException e) {
                  System.out.println("Can't handle input as short");
               } finally {
                  break;
               }
            case 2:  // Integer
               try {
                  System.out.print("Enter integer value: ");
                  int s1 = Integer.parseInt(stdin.readLine());
                  result = intToBytes(s1);
                  int s2 = bytesToInt(result);
                  if (s1 == s2) status = " OK "; else status = " FAIL ";
                  System.out.println("Retrieved " + s2 + status);
               } catch (NumberFormatException e) {
                  System.out.println("Can't handle input as integer");
               } finally {
                  break;
               }

            case 3:  // Long
               try {
                  System.out.print("Enter long value: ");
                  long s1 = Long.parseLong(stdin.readLine());
                  result = longToBytes(s1);
                  long s2 = bytesToLong(result);
                  if (s1 == s2) status = " OK "; else status = " FAIL ";
                  System.out.println("Retrieved " + s2 + status);
               } catch (NumberFormatException e) {
                  System.out.println("Can't handle input as long");
               } finally {
                  break;
               }

            case 4: // Packed Decimal
               try {
                  System.out.print("Enter packed decimal value: ");
                  String s1 = stdin.readLine();
                  System.out.print("Enter # of bytes to store in: ");
                  int blen = Integer.parseInt(stdin.readLine());
                  result = stringToPackedBytes(s1, blen);
                  String s2 = packedBytesToString(result);
                  if (s1.equals(s2)) status = " OK "; else status = " FAIL ";
                  System.out.println("Retrieved " + s2 + status);
               } catch (NumberFormatException e) {
                  System.out.println("Can't handle input as packed decimal");
               } catch (Exception e2) {
                  System.out.println(e2);
               } finally {
                  break;
               }

            case 5: // Float as IEEE Real
            case 6: // Float as HP Real
               if (function == 5) format = IEEE_FORMAT;
               if (function == 6) format = HP_FORMAT;
               try {
                  System.out.print("Enter float value: ");
                  float s1 = Float.valueOf(stdin.readLine()).floatValue();
                  result = floatToBytes(s1, format);
                  float s2 = bytesToFloat(result, format);
                  if (s1 == s2) status = " OK "; else status = " FAIL ";
                  System.out.println("Retrieved " + s2 + status);
               } catch (NumberFormatException e) {
                  System.out.println("Can't handle input as float");
               } finally {
                  break;
               }

            case 7: // Double as IEEE Real
            case 8: // Double as HP Real
               if (function == 7) format = IEEE_FORMAT;
               if (function == 8) format = HP_FORMAT;
               try {
                  System.out.print("Enter double value: ");
                  double s1 = Double.valueOf(stdin.readLine()).doubleValue();
//System.out.println(Long.toHexString(Double.doubleToLongBits(s1)));
                  result = doubleToBytes(s1, format);
                  double s2 = bytesToDouble(result, format);
//System.out.println(Long.toHexString(Double.doubleToLongBits(s2)));
                  if (s1 == s2) status = " OK "; else status = " FAIL ";
                  System.out.println("Retrieved " + s2 + status);
               } catch (NumberFormatException e) {
                  System.out.println("Can't handle input as double");
               } finally {
                  break;
               }

            default:
               System.out.println("Invalid function selection.  Try again");
         } // switch
      } // infinite loop
   } // main
}
